from enum import Enum
from typing import Dict

from boa3.neo.vm.type.Integer import Integer


class Opcode(bytes, Enum):

    # region Constants

    PUSHINT8 = b'\x00'
    PUSHINT16 = b'\x01'
    PUSHINT32 = b'\x02'
    PUSHINT64 = b'\x03'
    PUSHINT128 = b'\x04'
    PUSHINT256 = b'\x05'

    # Convert the next four bytes to an address, and push the address onto the stack.
    PUSHA = b'\x0A'
    # The item null is pushed onto the stack.
    PUSHNULL = b'\x0B'
    # The next byte contains the number of bytes to be pushed onto the stack.
    PUSHDATA1 = b'\x0C'
    # The next two bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA2 = b'\x0D'
    # The next four bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA4 = b'\x0E'

    @staticmethod
    def get_literal_push(integer: int):
        """
        Gets the push opcode to the respective integer

        :param integer: value that will be pushed
        :return: the respective opcode
        :rtype: Opcode or None
        """
        if -1 <= integer <= 16:
            opcode_value: int = Integer.from_bytes(Opcode.PUSH0) + integer
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            return None

    # The number -1 is pushed onto the stack.
    PUSHM1 = b'\x0F'
    # The number 0 is pushed onto the stack.
    PUSH0 = b'\x10'
    # The number 1 is pushed onto the stack.
    PUSH1 = b'\x11'
    # The number 2 is pushed onto the stack.
    PUSH2 = b'\x12'
    # The number 3 is pushed onto the stack.
    PUSH3 = b'\x13'
    # The number 4 is pushed onto the stack.
    PUSH4 = b'\x14'
    # The number 5 is pushed onto the stack.
    PUSH5 = b'\x15'
    # The number 6 is pushed onto the stack.
    PUSH6 = b'\x16'
    # The number 7 is pushed onto the stack.
    PUSH7 = b'\x17'
    # The number 8 is pushed onto the stack.
    PUSH8 = b'\x18'
    # The number 9 is pushed onto the stack.
    PUSH9 = b'\x19'
    # The number 10 is pushed onto the stack.
    PUSH10 = b'\x1A'
    # The number 11 is pushed onto the stack.
    PUSH11 = b'\x1B'
    # The number 12 is pushed onto the stack.
    PUSH12 = b'\x1C'
    # The number 13 is pushed onto the stack.
    PUSH13 = b'\x1D'
    # The number 14 is pushed onto the stack.
    PUSH14 = b'\x1E'
    # The number 15 is pushed onto the stack.
    PUSH15 = b'\x1F'
    # The number 16 is pushed onto the stack.
    PUSH16 = b'\x20'

    # endregion

    # region Flow control

    def has_larger_opcode(self) -> bool:
        return self in self.__larger_opcode

    def get_larger_opcode(self):
        """
        Gets the large opcode to the standard opcode

        :return: the respective opcode
        :rtype: Opcode or None
        """
        opcode_map = self.__larger_opcode
        if self in opcode_map:
            return opcode_map[self]
        elif self in opcode_map.values():
            return self
        else:
            return None

    @property
    def __larger_opcode(self) -> Dict[bytes, bytes]:
        """
        A map of each opcode with its larger equivalent

        :return: a dictionary that maps each opcode to its larger equivalent.
        """
        opcodes = {
            self.JMP: self.JMP_L,
            self.JMPIF: self.JMPIF_L,
            self.JMPIFNOT: self.JMPIFNOT_L,
            self.JMPEQ: self.JMPEQ_L,
            self.JMPNE: self.JMPNE_L,
            self.JMPGT: self.JMPGT_L,
            self.JMPGE: self.JMPGE_L,
            self.JMPLT: self.JMPLT_L,
            self.JMPLE: self.JMPLE_L,
            self.CALL: self.CALL_L,
            self.TRY: self.TRY_L,
            self.ENDTRY: self.ENDTRY_L
        }
        return opcodes

    # The NOP operation does nothing. It is intended to fill in space if opcodes are patched.
    NOP = b'\x21'
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 1-byte
    # signed offset from the beginning of the current instruction.
    JMP = b'\x22'
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 4-bytes
    # signed offset from the beginning of the current instruction.
    JMP_L = b'\x23'
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPIF = b'\x24'
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPIF_L = b'\x25'
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 1-byte signed offset from the beginning of the current
    # instruction.
    JMPIFNOT = b'\x26'
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 4-bytes signed offset from the beginning of the current
    # instruction.
    JMPIFNOT_L = b'\x27'
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 1-byte signed offset from the beginning of the current instruction.
    JMPEQ = b'\x28'
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 4-bytes signed offset from the beginning of the current instruction.
    JMPEQ_L = b'\x29'
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 1-byte signed offset from the beginning of the current instruction.
    JMPNE = b'\x2A'
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 4-bytes signed offset from the beginning of the current instruction.
    JMPNE_L = b'\x2B'
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPGT = b'\x2C'
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPGT_L = b'\x2D'
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The 
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction. 
    JMPGE = b'\x2E'
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The 
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction. 
    JMPGE_L = b'\x2F'
    # Transfers control to a target instruction if the first value is less than the second value. The target 
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction. 
    JMPLT = b'\x30'
    # Transfers control to a target instruction if the first value is less than the second value. The target 
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction. 
    JMPLT_L = b'\x31'
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The 
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction. 
    JMPLE = b'\x32'
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The 
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction. 
    JMPLE_L = b'\x33'

    # Calls the function at the target address which is represented as a 1-byte signed offset from the beginning of 
    # the current instruction. 
    CALL = b'\x34'
    # Calls the function at the target address which is represented as a 4-bytes signed offset from the beginning of 
    # the current instruction. 
    CALL_L = b'\x35'
    # Pop the address of a function from the stack, and call the function.
    CALLA = b'\x36'
    # It turns the vm state to FAULT immediately, and cannot be caught.
    ABORT = b'\x37'
    # Pop the top value of the stack, if it false, then exit vm execution and set vm state to FAULT.
    ASSERT = b'\x38'
    # Pop the top value of the stack, and throw it.
    THROW = b'\x3A'
    # TRY CatchOffset(sbyte) FinallyOffset(sbyte). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY = b'\x3B'
    # TRY_L CatchOffset(int) FinallyOffset(int). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY_L = b'\x3C'
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 1-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY = b'\x3D'
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 4-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY_L = b'\x3E'
    # End finally, If no exception happen or be catched, vm will jump to the target instruction of ENDTRY/ENDTRY_L.
    # Otherwise vm will rethrow the exception to upper layer.
    ENDFINALLY = b'\x3F'
    # Returns from the current method.
    RET = b'\x40'
    # Calls to an interop service.
    SYSCALL = b'\x41'

    def has_target(self) -> bool:
        return self.JMP <= self <= self.CALL_L or self.ENDTRY <= self <= self.ENDFINALLY

    # endregion

    # region Stack

    # Puts the number of stack items onto the stack.
    DEPTH = b'\x43'
    # Removes the top stack item.
    DROP = b'\x45'
    # Removes the second-to-top stack item.
    NIP = b'\x46'
    # The item n back in the main stack is removed.
    XDROP = b'\x48'
    # Clear the stack
    CLEAR = b'\x49'

    @staticmethod
    def get_drop(position: int):
        """
        Gets the opcode to remove the item n back in the stack

        :param position: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        duplicate_item = {
            1: Opcode.DROP,
            2: Opcode.NIP
        }

        if position > 0:
            if position in duplicate_item:
                return duplicate_item[position]
            else:
                return Opcode.XDROP

    # Duplicates the top stack item.
    DUP = b'\x4A'
    # Copies the second-to-top stack item to the top.
    OVER = b'\x4B'
    # The item n back in the stack is copied to the top.
    PICK = b'\x4D'

    @staticmethod
    def get_dup(position: int):
        """
        Gets the opcode to duplicate the item n back in the stack

        :param position: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        duplicate_item = {
            1: Opcode.DUP,
            2: Opcode.OVER
        }

        if position >= 0:
            if position in duplicate_item:
                return duplicate_item[position]
            else:
                return Opcode.PICK

    # The item at the top of the stack is copied and inserted before the second-to-top item.
    TUCK = b'\x4E'
    # The top two items on the stack are swapped.
    SWAP = b'\x50'
    # The top three items on the stack are rotated to the left.
    ROT = b'\x51'
    # The item n back in the stack is moved to the top.
    ROLL = b'\x52'
    # Reverse the order of the top 3 items on the stack.
    REVERSE3 = b'\x53'
    # Reverse the order of the top 4 items on the stack.
    REVERSE4 = b'\x54'
    # Pop the number N on the stack, and reverse the order of the top N items on the stack.
    REVERSEN = b'\x55'

    @staticmethod
    def get_reverse(no_stack_items: int):
        """
        Gets the opcode to reverse n items on the stack

        :param no_stack_items: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        reverse_stack = {
            2: Opcode.SWAP,
            3: Opcode.REVERSE3,
            4: Opcode.REVERSE4
        }

        if no_stack_items >= 0:
            if no_stack_items in reverse_stack:
                return reverse_stack[no_stack_items]
            else:
                return Opcode.REVERSEN

    # endregion

    # region Slot

    # Initialize the static field list for the current execution context.
    INITSSLOT = b'\x56'
    # Initialize the argument slot and the local variable list for the current execution context.
    INITSLOT = b'\x57'

    @staticmethod
    def get_store(index: int, local: bool, is_arg: bool = False) -> bytes:
        """
        Gets the opcode to store the variable

        :param index: index of the variable
        :param local: identifies if the variable is local or global
        :param is_arg: identifies if the variable is an argument of a function. False if local is False.
        :return: the respective opcode
        :rtype: Opcode
        """
        if not local:
            is_arg = False

        if 0 <= index <= 6:
            if is_arg:
                opcode_value: int = Integer.from_bytes(Opcode.STARG0) + index
            elif local:
                opcode_value: int = Integer.from_bytes(Opcode.STLOC0) + index
            else:
                opcode_value: int = Integer.from_bytes(Opcode.STSFLD0) + index
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            if is_arg:
                return Opcode.STARG
            elif local:
                return Opcode.STLOC
            else:
                return Opcode.STSFLD

    @staticmethod
    def get_load(index: int, local: bool, is_arg: bool = False):
        """
        Gets the opcode to load the variable

        :param index: index of the variable
        :param local: identifies if the variable is local or global
        :param is_arg: identifies if the variable is an argument of a function. False if local is False.
        :return: the respective opcode
        :rtype: Opcode
        """
        if not local:
            is_arg = False

        if 0 <= index <= 6:
            if is_arg:
                opcode_value: int = Integer.from_bytes(Opcode.LDARG0) + index
            elif local:
                opcode_value: int = Integer.from_bytes(Opcode.LDLOC0) + index
            else:
                opcode_value: int = Integer.from_bytes(Opcode.LDSFLD0) + index
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            if is_arg:
                return Opcode.LDARG
            elif local:
                return Opcode.LDLOC
            else:
                return Opcode.LDSFLD

    @property
    def is_load_slot(self) -> bool:
        return (self.LDSFLD0 <= self <= self.LDSFLD
                or self.LDLOC0 <= self <= self.LDLOC
                or self.LDARG0 <= self <= self.LDARG)

    @staticmethod
    def get_store_from_load(load_opcode):
        """
        Gets the store slot opcode equivalent to the given load slot opcode.

        :param load_opcode: load opcode
        :type load_opcode: Opcode
        :return: equivalent store opcode if the given opcode is a load slot. Otherwise, returns None
        :rtype: Opcode or None
        """
        if load_opcode.is_load_slot:
            opcode_value: int = Integer.from_bytes(load_opcode) + 8
            return Opcode(Integer(opcode_value).to_byte_array())

    # Loads the static field at index 0 onto the evaluation stack.
    LDSFLD0 = b'\x58'
    # Loads the static field at index 1 onto the evaluation stack.
    LDSFLD1 = b'\x59'
    # Loads the static field at index 2 onto the evaluation stack.
    LDSFLD2 = b'\x5A'
    # Loads the static field at index 3 onto the evaluation stack.
    LDSFLD3 = b'\x5B'
    # Loads the static field at index 4 onto the evaluation stack.
    LDSFLD4 = b'\x5C'
    # Loads the static field at index 5 onto the evaluation stack.
    LDSFLD5 = b'\x5D'
    # Loads the static field at index 6 onto the evaluation stack.
    LDSFLD6 = b'\x5E'
    # Loads the static field at a specified index onto the evaluation stack. The index is represented as a 1-byte 
    # unsigned integer. 
    LDSFLD = b'\x5F'
    # Stores the value on top of the evaluation stack in the static field list at index 0.
    STSFLD0 = b'\x60'
    # Stores the value on top of the evaluation stack in the static field list at index 1.
    STSFLD1 = b'\x61'
    # Stores the value on top of the evaluation stack in the static field list at index 2.
    STSFLD2 = b'\x62'
    # Stores the value on top of the evaluation stack in the static field list at index 3.
    STSFLD3 = b'\x63'
    # Stores the value on top of the evaluation stack in the static field list at index 4.
    STSFLD4 = b'\x64'
    # Stores the value on top of the evaluation stack in the static field list at index 5.
    STSFLD5 = b'\x65'
    # Stores the value on top of the evaluation stack in the static field list at index 6.
    STSFLD6 = b'\x66'
    # Stores the value on top of the evaluation stack in the static field list at a specified index. The index is 
    # represented as a 1-byte unsigned integer. 
    STSFLD = b'\x67'
    # Loads the local variable at index 0 onto the evaluation stack.
    LDLOC0 = b'\x68'
    # Loads the local variable at index 1 onto the evaluation stack.
    LDLOC1 = b'\x69'
    # Loads the local variable at index 2 onto the evaluation stack.
    LDLOC2 = b'\x6A'
    # Loads the local variable at index 3 onto the evaluation stack.
    LDLOC3 = b'\x6B'
    # Loads the local variable at index 4 onto the evaluation stack.
    LDLOC4 = b'\x6C'
    # Loads the local variable at index 5 onto the evaluation stack.
    LDLOC5 = b'\x6D'
    # Loads the local variable at index 6 onto the evaluation stack.
    LDLOC6 = b'\x6E'
    # Loads the local variable at a specified index onto the evaluation stack. The index is represented as a 1-byte 
    # unsigned integer. 
    LDLOC = b'\x6F'
    # Stores the value on top of the evaluation stack in the local variable list at index 0.
    STLOC0 = b'\x70'
    # Stores the value on top of the evaluation stack in the local variable list at index 1.
    STLOC1 = b'\x71'
    # Stores the value on top of the evaluation stack in the local variable list at index 2.
    STLOC2 = b'\x72'
    # Stores the value on top of the evaluation stack in the local variable list at index 3.
    STLOC3 = b'\x73'
    # Stores the value on top of the evaluation stack in the local variable list at index 4.
    STLOC4 = b'\x74'
    # Stores the value on top of the evaluation stack in the local variable list at index 5.
    STLOC5 = b'\x75'
    # Stores the value on top of the evaluation stack in the local variable list at index 6.
    STLOC6 = b'\x76'
    # Stores the value on top of the evaluation stack in the local variable list at a specified index. The index is 
    # represented as a 1-byte unsigned integer. 
    STLOC = b'\x77'
    # Loads the argument at index 0 onto the evaluation stack.
    LDARG0 = b'\x78'
    # Loads the argument at index 1 onto the evaluation stack.
    LDARG1 = b'\x79'
    # Loads the argument at index 2 onto the evaluation stack.
    LDARG2 = b'\x7A'
    # Loads the argument at index 3 onto the evaluation stack.
    LDARG3 = b'\x7B'
    # Loads the argument at index 4 onto the evaluation stack.
    LDARG4 = b'\x7C'
    # Loads the argument at index 5 onto the evaluation stack.
    LDARG5 = b'\x7D'
    # Loads the argument at index 6 onto the evaluation stack.
    LDARG6 = b'\x7E'
    # Loads the argument at a specified index onto the evaluation stack. The index is represented as a 1-byte 
    # unsigned integer. 
    LDARG = b'\x7F'
    # Stores the value on top of the evaluation stack in the argument slot at index 0.
    STARG0 = b'\x80'
    # Stores the value on top of the evaluation stack in the argument slot at index 1.
    STARG1 = b'\x81'
    # Stores the value on top of the evaluation stack in the argument slot at index 2.
    STARG2 = b'\x82'
    # Stores the value on top of the evaluation stack in the argument slot at index 3.
    STARG3 = b'\x83'
    # Stores the value on top of the evaluation stack in the argument slot at index 4.
    STARG4 = b'\x84'
    # Stores the value on top of the evaluation stack in the argument slot at index 5.
    STARG5 = b'\x85'
    # Stores the value on top of the evaluation stack in the argument slot at index 6.
    STARG6 = b'\x86'
    # Stores the value on top of the evaluation stack in the argument slot at a specified index. The index is 
    # represented as a 1-byte unsigned integer. 
    STARG = b'\x87'

    # endregion

    # region Splice

    NEWBUFFER = b'\x88',
    MEMCPY = b'\x89',
    # Concatenates two strings.
    CAT = b'\x8B'
    # Returns a section of a string.
    SUBSTR = b'\x8C'
    # Keeps only characters left of the specified point in a string.
    LEFT = b'\x8D'
    # Keeps only characters right of the specified point in a string.
    RIGHT = b'\x8E'

    # endregion

    # region Bitwise logic

    # Flips all of the bits in the input.
    INVERT = b'\x90'
    # Boolean and between each bit in the inputs.
    AND = b'\x91'
    # Boolean or between each bit in the inputs.
    OR = b'\x92'
    # Boolean exclusive or between each bit in the inputs.
    XOR = b'\x93'
    # Returns 1 if the inputs are exactly equal, 0 otherwise.
    EQUAL = b'\x97'
    # Returns 1 if the inputs are not equal, 0 otherwise.
    NOTEQUAL = b'\x98'

    # endregion

    # region Arithmetic

    # Puts the sign of top stack item on top of the main stack. If value is negative, put -1; if positive, 
    # put 1; if value is zero, put 0. 
    SIGN = b'\x99'
    # The input is made positive.
    ABS = b'\x9A'
    # The sign of the input is flipped.
    NEGATE = b'\x9B'
    # 1 is added to the input.
    INC = b'\x9C'
    # 1 is subtracted from the input.
    DEC = b'\x9D'
    # a is added to b.
    ADD = b'\x9E'
    # b is subtracted from a.
    SUB = b'\x9F'
    # a is multiplied by b.
    MUL = b'\xA0'
    # a is divided by b.
    DIV = b'\xA1'
    # Returns the remainder after dividing a by b.
    MOD = b'\xA2'
    # Shifts a left b bits, preserving sign.
    SHL = b'\xA8'
    # Shifts a right b bits, preserving sign.
    SHR = b'\xA9'
    # If the input is 0 or 1, it is flipped. Otherwise the output will be 0.
    NOT = b'\xAA'
    # If both a and b are not 0, the output is 1. Otherwise 0.
    BOOLAND = b'\xAB'
    # If a or b is not 0, the output is 1. Otherwise 0.
    BOOLOR = b'\xAC'
    # Returns 0 if the input is 0. 1 otherwise.
    NZ = b'\xB1'
    # Returns 1 if the numbers are equal, 0 otherwise.
    NUMEQUAL = b'\xB3'
    # Returns 1 if the numbers are not equal, 0 otherwise.
    NUMNOTEQUAL = b'\xB4'
    # Returns 1 if a is less than b, 0 otherwise.
    LT = b'\xB5'
    # Returns 1 if a is less than or equal to b, 0 otherwise.
    LE = b'\xB6'
    # Returns 1 if a is greater than b, 0 otherwise.
    GT = b'\xB7'
    # Returns 1 if a is greater than or equal to b, 0 otherwise.
    GE = b'\xB8'
    # Returns the smaller of a and b.
    MIN = b'\xB9'
    # Returns the larger of a and b.
    MAX = b'\xBA'
    # Returns 1 if x is within the specified range (left-inclusive), 0 otherwise.
    WITHIN = b'\xBB'

    # endregion

    # region Compound-type

    # A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized array
    # and this array is put on top of the main stack. 
    PACK = b'\xC0'
    # An array is removed from top of the main stack. Its elements are put on top of the main stack (in reverse 
    # order) and the array size is also put on main stack. 
    UNPACK = b'\xC1'
    # An empty array (with size 0) is put on top of the main stack.
    NEWARRAY0 = b'\xC2'
    # A value n is taken from top of main stack. A null-filled array with size n is put on top of the main stack.
    NEWARRAY = b'\xC3'
    # A value n is taken from top of main stack. An array of type T with size n is put on top of the main stack.
    NEWARRAY_T = b'\xC4'
    # An empty struct (with size 0) is put on top of the main stack.
    NEWSTRUCT0 = b'\xC5'
    # A value n is taken from top of main stack. A zero-filled struct with size n is put on top of the main stack.
    NEWSTRUCT = b'\xC6'
    # A Map is created and put on top of the main stack.
    NEWMAP = b'\xC8'
    # An array is removed from top of the main stack. Its size is put on top of the main stack.
    SIZE = b'\xCA'
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Puts True on top of
    # main stack if array[n] (or map[n]) exist, and False otherwise. 
    HASKEY = b'\xCB'
    # A map is taken from top of the main stack. The keys of this map are put on top of the main stack.
    KEYS = b'\xCC'
    # A map is taken from top of the main stack. The values of this map are put on top of the main stack.
    VALUES = b'\xCD'
    # An input index n (or key) and an array (or map) are taken from main stack. Element array[n] (or map[n]) is put 
    # on top of the main stack. 
    PICKITEM = b'\xCE'
    # The item on top of main stack is removed and appended to the second item on top of the main stack.
    APPEND = b'\xCF'
    # A value v, index n (or key) and an array (or map) are taken from main stack. Attribution array[n]=v
    # (or map[n]=v) is performed.
    SETITEM = b'\xD0'
    # An array is removed from the top of the main stack and its elements are reversed.
    REVERSEITEMS = b'\xD1'
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Element array[n]
    # (or map[n]) is removed.
    REMOVE = b'\xD2'
    # Remove all the items from the compound-type.
    CLEARITEMS = b'\xD3'

    # endregion

    # region Types

    # Returns true if the input is null. Returns false otherwise.
    ISNULL = b'\xD8'
    # Returns true if the top item is of the specified type.
    ISTYPE = b'\xD9'
    # Converts the top item to the specified type.
    CONVERT = b'\xDB'

    # endregion



import ast
from typing import Dict, List, Optional

from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.type import Type
from boa3.model.variable import Variable


class Event(Callable, IdentifiedSymbol):
    def __init__(self, event_id: str, args: Dict[str, Variable] = None, defaults: List[ast.AST] = None,
                 origin_node: Optional[ast.AST] = None):
        super().__init__(args, defaults, Type.none, True, origin_node)

        self.name: str = event_id
        self._identifier: str = self.name

    @property
    def shadowing_name(self) -> str:
        return 'event'

    @property
    def identifier(self) -> str:
        return self.name



from typing import Any

from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItemType import StackItemType


class BytesType(SequenceType, PrimitiveType):
    """
    A class used to represent Python bytes type
    """

    def __init__(self):
        identifier = 'bytes'
        from boa3.model.type.primitive.inttype import IntType
        values_type = [IntType()]
        super().__init__(identifier, values_type)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.ByteArray

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Buffer

    @property
    def default_value(self) -> Any:
        return bytes()

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.model.type.type import Type
        return Type.int

    @classmethod
    def build(cls, value: Any):
        from boa3.model.type.type import Type
        return Type.bytes

    @classmethod
    def build_collection(cls, *value_type: IType):
        return cls.build(value_type)

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is bytes or isinstance(value, BytesType)

    @property
    def can_reassign_values(self) -> bool:
        return False



from typing import Any

from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.type.primitive.bytestype import BytesType


class ByteArrayType(BytesType, MutableSequenceType):
    """
    A class used to represent Python list type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'bytearray'

    @property
    def default_value(self) -> Any:
        return bytearray()

    @classmethod
    def build(cls, value: Any):
        from boa3.model.type.type import Type
        return Type.bytearray

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) in [bytearray, ByteArrayType]

    @property
    def can_reassign_values(self) -> bool:
        return True



import mmh3  # type: ignore

from boa3.neo3.core import serialization, types, Size as s


class StorageKey(serialization.ISerializable):
    def __init__(self, contract: types.UInt160 = None, key: bytes = None):
        self.contract = contract if contract else types.UInt160.zero()
        self.key = key if key else b''

    def __len__(self):
        # TODO: see if there is a cleaner way of doing this
        with serialization.BinaryWriter() as bw:
            bw.write_bytes_with_grouping(self.key, 16)
            key_len = len(bw._stream.getvalue())
        return s.uint160 + key_len

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.contract == other.contract and self.key == other.key

    def __hash__(self):
        return hash(self.contract) + mmh3.hash(self.key, seed=0, signed=False)

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> [{self.contract}] {self.key}"

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_serializable(self.contract)
        writer.write_bytes_with_grouping(self.key, group_size=16)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.contract = reader.read_serializable(types.UInt160)
        self.key = reader.read_bytes_with_grouping(group_size=16)

    # TODO: Remove when we can conclude with certainty it is no longer needed.
    # To be validated on neo-preview2 compatible release
    # @staticmethod
    # def create_grouped_prefix(key_prefix: bytes, group_size) -> bytes:
    #     if group_size >= 255:
    #         raise ValueError("group_size must be < 254")
    #
    #     output = b''
    #     index = 0
    #     remainder = len(key_prefix)
    #     while remainder >= group_size:
    #         output += key_prefix[index:index + group_size]
    #         output += bytes([group_size])
    #         index += group_size
    #         remainder -= group_size
    #
    #     if remainder > 0:
    #         output += key_prefix[index:]
    #     return output



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ByteArrayMethod(IBuiltinMethod):

    def __init__(self, argument_type: IType = None):
        from boa3.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'bytearray'
        args: Dict[str, Variable] = {'object': Variable(argument_type)}
        super().__init__(identifier, args, return_type=Type.bytearray)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.model.type.type import Type
        # TODO: change when building bytearray given size is implemented
        return (isinstance(param_type, SequenceType)
                and (param_type is Type.str
                     or isinstance(param_type.value_type, type(Type.int))
                     ))

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def is_supported(self) -> bool:
        # TODO: change when building bytearray from string and int iterators are implemented
        from boa3.model.type.type import Type
        return self._arg_object.type is Type.bytes

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any):
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ByteArrayMethod(value)
        return super().build(value)



from abc import ABC

from boa3.model.type.itype import IType


class PrimitiveType(IType, ABC):
    """
    An interface for primitive types
    """

    def __init__(self, identifier: str):
        super().__init__(identifier)



from typing import Optional, Any

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.type.AbiType import AbiType


class CreateEventMethod(IBuiltinMethod):
    def __init__(self):
        import ast
        from boa3.model.type.type import Type
        identifier = 'CreateNewEvent'
        args = {
            'arguments': Variable(Type.list.build(Type.tuple)),
            'event_name': Variable(Type.str)
        }
        event_name_default = ast.parse("'{0}'".format(Type.str.default_value)
                                       ).body[0].value
        super().__init__(identifier, args, defaults=[event_name_default], return_type=EventType)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == len(self.args)

    @property
    def _args_on_stack(self) -> int:
        return 0

    @property
    def _body(self) -> Optional[str]:
        return None


class __EventType(IType):
    """
    A class used to represent an Neo event
    """

    def __init__(self):
        identifier = 'event'
        super().__init__(identifier)

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Void

    @classmethod
    def build(cls, value: Any):
        return EventType

    @classmethod
    def _is_type_of(cls, value: Any):
        return value is EventType


EventType = __EventType()



from abc import ABC

from boa3.model.symbol import ISymbol


class IdentifiedSymbol(ISymbol, ABC):
    """
    A class to represent a symbol with a name identifier attached to it

    :return: the resulting type when the expression is evaluated
    """

    def __init__(self, identifier: str):
        self._identifier: str = identifier

    @property
    def identifier(self) -> str:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        return self._identifier



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, UnresolvedOperation
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestTuple(BoaTest):

    def test_tuple_int_values(self):
        path = '%s/boa3_test/test_sc/tuple_test/IntTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = (1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_string_values(self):
        path = '%s/boa3_test/test_sc/tuple_test/StrTuple.py' % self.dirname
        byte_input0 = String('1').to_bytes()
        byte_input1 = String('2').to_bytes()
        byte_input2 = String('3').to_bytes()

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = ('1', '2', '3')
            + Integer(len(byte_input2)).to_byte_array()
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSHDATA1
            + Integer(len(byte_input0)).to_byte_array()
            + byte_input0
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_bool_values(self):
        path = '%s/boa3_test/test_sc/tuple_test/BoolTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = (True, True, False)
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_variable_values(self):
        path = '%s/boa3_test/test_sc/tuple_test/VariableTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH1      # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2      # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH3      # c = 3
            + Opcode.STLOC2
            + Opcode.LDLOC2     # d = (a, b, c)
            + Opcode.LDLOC1
            + Opcode.LDLOC0
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_assign_empty_tuple(self):
        path = '%s/boa3_test/test_sc/tuple_test/EmptyTupleAssignment.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.NEWARRAY0
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_get_value(self):
        path = '%s/boa3_test/test_sc/tuple_test/GetValue.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_non_sequence_get_value(self):
        path = '%s/boa3_test/test_sc/tuple_test/MismatchedTypeGetValue.py' % self.dirname
        self.assertCompilerLogs(UnresolvedOperation, path)

    def test_tuple_set_value(self):
        path = '%s/boa3_test/test_sc/tuple_test/SetValue.py' % self.dirname
        self.assertCompilerLogs(UnresolvedOperation, path)

    def test_non_sequence_set_value(self):
        path = '%s/boa3_test/test_sc/tuple_test/MismatchedTypeSetValue.py' % self.dirname
        self.assertCompilerLogs(UnresolvedOperation, path)

    def test_tuple_index_mismatched_type(self):
        path = '%s/boa3_test/test_sc/tuple_test/MismatchedTypeTupleIndex.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_tuple_of_tuple(self):
        path = '%s/boa3_test/test_sc/tuple_test/TupleOfTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0][0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_nep5_main(self):
        path = '%s/boa3_test/test_sc/tuple_test/Nep5Main.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG1     # args[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[2:3]
            + Opcode.PUSH2
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PUSH3
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PUSH2      # get slice
                + Opcode.PICK
                + Opcode.SIZE
                + Opcode.MIN        # slice end
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingLiteralValues.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_with_variables(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH2      # a1 = 2
            + Opcode.STLOC0
            + Opcode.PUSH3      # a2 = 3
            + Opcode.STLOC1
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.LDLOC2     # return a[a1:a2]
            + Opcode.LDLOC0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.LDLOC1
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PUSH2      # get slice
                + Opcode.PICK
                + Opcode.SIZE
                + Opcode.MIN        # slice end
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingVariableValues.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_negative_start(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[-4:]
            + Opcode.DUP
            + Opcode.SIZE       # slice end
            + Opcode.PUSH4
            + Opcode.NEGATE
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.SWAP       # get slice
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingNegativeStart.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_negative_end(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[:-4]
            + Opcode.DUP
            + Opcode.SIZE
            + Opcode.PUSH4          # slice end
            + Opcode.NEGATE
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.SWAP
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingNegativeEnd.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_start_omitted(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[:3]
            + Opcode.PUSH3          # slice end
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PUSH0
                + Opcode.SWAP
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingStartOmitted.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_omitted(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[:]
            + Opcode.UNPACK
            + Opcode.PACK
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingOmitted.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_end_omitted(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = [0, 1, 2, 3, 4, 5]
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[2:]
            + Opcode.DUP
            + Opcode.SIZE       # slice end
            + Opcode.PUSH2
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.SWAP       # get slice
                + Opcode.NEWARRAY0  # slice
                + Opcode.PUSH2
                + Opcode.PICK       # index
                + Opcode.JMP        # while index < end
                + Integer(32).to_byte_array(min_length=1)
                + Opcode.DUP            # if index >= slice start
                + Opcode.PUSH4
                + Opcode.PICK
                + Opcode.GE
                + Opcode.JMPIFNOT
                + Integer(25).to_byte_array(min_length=1)
                + Opcode.OVER               # slice.append(array[index])
                + Opcode.PUSH5
                + Opcode.PICK
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(signed=True, min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.INC            # index += 1
                + Opcode.DUP
                + Opcode.PUSH3
                + Opcode.PICK
                + Opcode.LT
                + Opcode.JMPIF          # end while index < slice end
                + Integer(-34).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.REVERSE4
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.DROP
            + Opcode.RET        # return
        )
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingEndOmitted.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_slicing_omitted_stride(self):
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingWithStride.py' % self.dirname
        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)

    def test_tuple_slicing_omitted_with_stride(self):
        path = '%s/boa3_test/test_sc/tuple_test/TupleSlicingOmittedWithStride.py' % self.dirname
        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)



from boa3.model.callable import Callable
from boa3.neo.vm.VMCode import VMCode


class CallCode(VMCode):
    """
    Represents a Neo VM function call code

    :ivar info: the opcode information of the code
    :ivar data: the data in bytes of the code. Empty byte array by default.
    """

    def __init__(self, target_callable: Callable):
        """
        :param target_callable: the calling method
        :type target_callable: Method
        """
        self._target_callable: Callable = target_callable
        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        super().__init__(OpcodeInfo.CALL)

    @property
    def target(self) -> VMCode:
        return self._target_callable.init_bytecode



from typing import Dict

from boa3.boa3 import Boa3
from boa3.compiler.compiler import Compiler
from boa3.exception.CompilerError import MismatchedTypes, NotSupportedOperation, UnresolvedReference
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestVariable(BoaTest):

    def test_declaration_with_type(self):
        path = '%s/boa3_test/test_sc/variable_test/DeclarationWithType.py' % self.dirname

        test_variable_id = 'a'
        test_method_id = 'Main'
        compiler = Compiler()

        expected_compiler_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        compiler_output = compiler.compile(path)
        self.assertEqual(expected_compiler_output, compiler_output)

        main_symbol_table: Dict[str, ISymbol] = self.get_compiler_analyser(compiler).symbol_table
        # the variable is local to a method, so it shouldn't be in the main symbol table
        self.assertFalse(test_variable_id in main_symbol_table)

        self.assertTrue(test_method_id in main_symbol_table)
        self.assertIsInstance(main_symbol_table[test_method_id], Method)
        method: Method = main_symbol_table[test_method_id]

        method_symbol_table: Dict[str, Variable] = method.symbols
        # the variable is local to this method, so it should be in the method symbol table
        self.assertTrue(test_variable_id in method_symbol_table)

    def test_declaration_without_type(self):
        path = '%s/boa3_test/test_sc/variable_test/DeclarationWithoutType.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_assignment_with_type(self):
        input = 'unit_test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # assignment value
            + Integer(len(byte_input)).to_byte_array(min_length=1)
            + byte_input
            + Opcode.STLOC0     # variable address
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/AssignmentWithType.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_assignment_without_type(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1      # assignment value
            + Opcode.STLOC0     # variable address
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/AssignmentWithoutType.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_argument_assignment(self):
        input = 'unit_test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT         # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.PUSHDATA1      # assignment value
            + Integer(len(byte_input)).to_byte_array(min_length=1)
            + byte_input
            + Opcode.STARG0         # variable address
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/ArgumentAssignment.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_assignments(self):
        path = '%s/boa3_test/test_sc/variable_test/MultipleAssignments.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_tuple_multiple_assignments(self):
        path = '%s/boa3_test/test_sc/variable_test/AssignmentWithTuples.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_many_assignments(self):
        expected_output = (
            Opcode.INITSLOT         # function signature
            + b'\x08'
            + b'\x00'
            + Opcode.PUSH0          # function body
            + Opcode.STLOC0
            + Opcode.PUSH1
            + Opcode.STLOC1
            + Opcode.PUSH2
            + Opcode.STLOC2
            + Opcode.PUSH3
            + Opcode.STLOC3
            + Opcode.PUSH4
            + Opcode.STLOC4
            + Opcode.PUSH5
            + Opcode.STLOC5
            + Opcode.PUSH6
            + Opcode.STLOC6
            + Opcode.PUSH7
            + Opcode.STLOC          # variable index greater than 6 uses another opcode
            + b'\x07'
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/ManyAssignments.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_return_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # variable address
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/ReturnArgument.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_return_local_var_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH1
            + Opcode.STLOC0
            + Opcode.LDLOC0     # variable address
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/ReturnLocalVariable.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_assign_local_with_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0     # variable address
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/AssignLocalWithArgument.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_using_undeclared_variable(self):
        path = '%s/boa3_test/test_sc/variable_test/UsingUndeclaredVariable.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_return_undeclared_variable(self):
        path = '%s/boa3_test/test_sc/variable_test/ReturnUndeclaredVariable.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_assign_value_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignValue.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_assign_binary_operation_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignBinOp.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_assign_unary_operation_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignUnOp.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_assign_mixed_operations_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignMixedOp.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_assign_sequence_get_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignSequenceGet.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_assign_sequence_set_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAssignSequenceSet.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_aug_assign_mismatched_type(self):
        path = '%s/boa3_test/test_sc/variable_test/MismatchedTypeAugAssign.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_global_declaration_with_assignment(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/GlobalDeclarationWithArgumentWrittenAfter.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_global_declaration_without_assignment(self):
        path = '%s/boa3_test/test_sc/variable_test/GlobalDeclarationWithoutAssignment.py' % self.dirname
        self.assertCompilerLogs(UnresolvedReference, path)

    def test_global_assignment_with_type(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentWithType.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_global_assignment_without_type(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentWithoutType.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_global_tuple_multiple_assignments(self):
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentWithTuples.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_many_global_assignments(self):
        expected_output = (
            Opcode.LDSFLD + b'\x07'
            + Opcode.LDSFLD6    # [a, b, c, d, e, f, g, h]
            + Opcode.LDSFLD5
            + Opcode.LDSFLD4
            + Opcode.LDSFLD3
            + Opcode.LDSFLD2
            + Opcode.LDSFLD1
            + Opcode.LDSFLD0
            + Opcode.PUSH8
            + Opcode.PACK       # return [a, b, c, d, e, f, g, h]
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x08'           # number of globals
            + Opcode.PUSH0      # a = 0
            + Opcode.STSFLD0
            + Opcode.PUSH1      # b = 1
            + Opcode.STSFLD1
            + Opcode.PUSH2      # c = 2
            + Opcode.STSFLD2
            + Opcode.PUSH3      # d = 3
            + Opcode.STSFLD3
            + Opcode.PUSH4      # e = 4
            + Opcode.STSFLD4
            + Opcode.PUSH5      # f = 5
            + Opcode.STSFLD5
            + Opcode.PUSH6      # g = 6
            + Opcode.STSFLD6
            + Opcode.PUSH7      # h = 7
            + Opcode.STSFLD + b'\x07'   # variable index greater than 6 uses another opcode
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/variable_test/ManyGlobalAssignments.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_global_assignment_between_functions(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.LDSFLD1
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x02'           # number of globals
            + Opcode.PUSH10     # a = 10
            + Opcode.STSFLD0
            + Opcode.PUSH5      # b = 5
            + Opcode.STSFLD1
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentBetweenFunctions.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_get_global_variable_value_written_after(self):
        expected_output = (
            Opcode.LDSFLD + b'\x07'
            + Opcode.LDSFLD6    # [a, b, c, d, e, f, g, h]
            + Opcode.LDSFLD5
            + Opcode.LDSFLD4
            + Opcode.LDSFLD3
            + Opcode.LDSFLD2
            + Opcode.LDSFLD1
            + Opcode.LDSFLD0
            + Opcode.PUSH8
            + Opcode.PACK       # return [a, b, c, d, e, f, g, h]
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x08'           # number of globals
            + Opcode.PUSH0      # a = 0
            + Opcode.STSFLD0
            + Opcode.PUSH1      # b = 1
            + Opcode.STSFLD1
            + Opcode.PUSH2      # c = 2
            + Opcode.STSFLD2
            + Opcode.PUSH3      # d = 3
            + Opcode.STSFLD3
            + Opcode.PUSH4      # e = 4
            + Opcode.STSFLD4
            + Opcode.PUSH5      # f = 5
            + Opcode.STSFLD5
            + Opcode.PUSH6      # g = 6
            + Opcode.STSFLD6
            + Opcode.PUSH7      # h = 7
            + Opcode.STSFLD + b'\x07'   # variable index greater than 6 uses another opcode
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/GetGlobalValueWrittenAfter.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assign_local_shadowing_global_with_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # b = a  // this b is not the global b
            + Opcode.STLOC0
            + Opcode.LDLOC0     # variable address
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH0      # b = 0
            + Opcode.STSFLD0
            + Opcode.RET
        )
        path = '%s/boa3_test/test_sc/variable_test/AssignLocalWithArgumentShadowingGlobal.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assign_global_in_function(self):
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentInFunctionWithArgument.py' % self.dirname

        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)



from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.expression import IExpression


class MetadataDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = 'metadata'
        super().__init__(identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == len(self.args)



from __future__ import annotations

import abc
from enum import IntEnum

from events import Events  # type: ignore

msgrouter = Events()

# :noindex:


class Size(IntEnum):
    """
    Explicit bytes of memory consumed
    """
    uint8 = 1
    uint16 = 2
    uint32 = 4
    uint64 = 8
    uint160 = 20
    uint256 = 32


class IClonable(abc.ABC):
    @abc.abstractmethod
    def clone(self):
        """
        Create a deep copy of self
        """

    def from_replica(self, replica):
        pass


class IJson(abc.ABC):
    @abc.abstractmethod
    def to_json(self) -> dict:
        pass

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json: dict):
        pass



from abc import ABC
from typing import Iterable, Optional, Union


class CompilerError(ABC, Exception):
    """
    An interface for compilation errors
    """

    def __init__(self, line: int, col: int):
        self.line: int = line
        self.col: int = col

    @property
    def message(self) -> str:
        message = '' if self._error_message is None else ' - ' + self._error_message
        return '{0}:{1}{2}'.format(self.line, self.col, message)

    @property
    def _error_message(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self.message

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.message == other.message


class IncorrectNumberOfOperands(CompilerError):
    """
    An error raised when an operation is used with the wrong number of operands
    """

    def __init__(self, line: int, col: int, expected_count: int, actual_count: int):
        self.expected = expected_count
        self.actual = actual_count
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Incorrect number of operands: expected '%s', got '%s' instead" % (self.expected, self.actual)


class InvalidType(CompilerError):
    """
    An error raised when a type that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        message = "Invalid type"
        if self.symbol_id is not None:
            message += ": '%s'" % self.symbol_id
        return message


class MetadataImplementationMissing(CompilerError):
    """
    An error raised when the metadata required functions aren't implemented
    """

    def __init__(self, line: int, col: int, symbol_id: str, metadata_attr_id: str):
        self.symbol_id: str = symbol_id
        self.metadata_attr_id: str = metadata_attr_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'{0}' requires '{1}' implementation".format(self.metadata_attr_id, self.symbol_id)


class MetadataIncorrectImplementation(CompilerError):
    """
    An error raised when a metadata required function is incorrectly implemented
    """
    from boa3.model.symbol import ISymbol

    def __init__(self, line: int, col: int, symbol_id: str, expected_symbol: ISymbol, actual_symbol: ISymbol):
        from boa3.model.symbol import ISymbol

        self.symbol_id: str = symbol_id
        self.expected_symbol: ISymbol = expected_symbol
        self.actual_symbol: ISymbol = actual_symbol
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return ("'{0}' is not correctly implemented. Expecting '{1}', got '{2}' instead"
                .format(self.symbol_id, self.expected_symbol, self.actual_symbol))


class MetadataInformationMissing(CompilerError):
    """
    An error raised when the metadata info doesn't match with the functions requirements
    """

    def __init__(self, line: int, col: int, symbol_id: str, metadata_attr_id: str):
        self.symbol_id: str = symbol_id
        self.metadata_attr_id: str = metadata_attr_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'{0}' requires '{1}' attribute, which is missing in the metadata".format(self.symbol_id, self.metadata_attr_id)


class MismatchedTypes(CompilerError):
    """
    An error raised when the evaluated and expected types are not the same
    """

    def __init__(self, line: int, col: int, expected_type_id: Union[str, Iterable[str]], actual_type_id: Union[str, Iterable[str]]):
        if isinstance(expected_type_id, str):
            expected_type_id = [expected_type_id]
        if isinstance(actual_type_id, str):
            actual_type_id = [actual_type_id]

        self.expected_types = expected_type_id
        self.actual_types = actual_type_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        expected_types = join_args(self.expected_types)
        actual_types = join_args(self.actual_types)
        return "Expected type '%s', got '%s' instead" % (expected_types, actual_types)


class MissingReturnStatement(CompilerError):
    """
    An error raised when a function with a return value is missing a return statement
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'%s': Missing return statement" % self.symbol_id


class NotSupportedOperation(CompilerError):
    """
    An error raised when an operation that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "The following operation is not supported: '%s'" % self.symbol_id


class UnexpectedArgument(CompilerError):
    """
    An error thrown when more arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unexpected argument"


class UnfilledArgument(CompilerError):
    """
    An error thrown when less arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int, param: str):
        self.param = param
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Parameter '%s' unfilled" % self.param


class UnresolvedReference(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference '%s'" % self.symbol_id

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._error_message == other._error_message


class UnresolvedOperation(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, type_id: str, operation_id: str):
        self.type_id = type_id
        self.operation_id = operation_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference: '%s' does not have a definition of '%s' operator" % (self.type_id, self.operation_id)


class TooManyReturns(CompilerError):
    """
    An error raised when a function returns a tuple
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Too many returns"


class TypeHintMissing(CompilerError):
    """
    An error raised when type hint cannot be found
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Type hint is missing for the symbol '%s'" % self.symbol_id


def join_args(iterable: Iterable[str]) -> str:
    return str.join("', '", iterable)



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, NotSupportedOperation, UnresolvedOperation
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest


class TestBytes(BoaTest):

    def test_bytes_literal_value(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = b'\x01\x02\x03'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesLiteral.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesGetValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_get_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH1
            + Opcode.NEGATE
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesGetValueNegativeIndex.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_set_value(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesSetValue.py' % self.dirname
        self.assertCompilerLogs(UnresolvedOperation, path)

    def test_bytes_clear(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesClear.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_bytes_reverse(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesReverse.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_bytes_to_int(self):
        data = b'\x01\x02'
        expected_output = (
            Opcode.PUSHDATA1    # b'\x01\x02'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # b'\x01\x02'.to_int()
            + Type.int.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesToInt.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_to_int_with_builtin(self):
        data = b'\x01\x02'
        expected_output = (
            Opcode.PUSHDATA1    # b'\x01\x02'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # bytes.to_int(b'\x01\x02')
            + Type.int.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesToIntWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_to_int_mismatched_types(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesToIntWithBuiltinMismatchedTypes.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_bytes_to_int_with_byte_array_builtin(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesToIntWithBytearrayBuiltin.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_bytes_to_str(self):
        data = b'abc'
        expected_output = (
            Opcode.PUSHDATA1    # b'abc'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # b'abc'.to_str()
            + Type.str.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesToStr.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_to_str_with_builtin(self):
        data = b'123'
        expected_output = (
            Opcode.PUSHDATA1    # b'123'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # bytes.to_str(b'123')
            + Type.str.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesToStrWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_bytes_to_str_mismatched_types(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytesToStrWithBuiltinMismatchedTypes.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_bytes_from_byte_array(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = a
            + Opcode.STLOC1
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytesFromBytearray.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayGetValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_get_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH1
            + Opcode.NEGATE
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayGetValueNegativeIndex.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_set_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0] = 0x01
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.LDARG0
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearraySetValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_set_value_negative_index(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[-1] = 0x01
            + Opcode.PUSH1
            + Opcode.NEGATE
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.LDARG0
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearraySetValueNegativeIndex.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_literal_value(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytearrayLiteral.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_byte_array_from_literal_bytes(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayFromLiteralBytes.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_from_variable_bytes(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = b'\x01\x02\x03'
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = bytearray(a)
            + Opcode.STLOC1
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayFromVariableBytes.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_string(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytearrayFromString.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_byte_array_append(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.append(4)
            + Opcode.PUSH4
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET        # return a
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayAppend.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_append_with_builtin(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # bytearray.append(a, 4)
            + Opcode.PUSH4
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET        # return a
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayAppendWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_append_mutable_sequence_with_builtin(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # MutableSequence.append(a, 4)
            + Opcode.PUSH4
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET        # return a
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayAppendWithMutableSequence.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_clear(self):
        data = b'\x01\x02\x03\x04'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03\x04')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.clear()
                + Opcode.DUP
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(9).to_byte_array(min_length=1)
                + Opcode.DROP
                + Opcode.PUSHDATA1
                + Integer(0).to_byte_array(min_length=1)
                + Opcode.CONVERT
                + Type.bytearray.stack_item
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CLEARITEMS
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET        # return a
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayClear.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_reverse(self):
        data = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = bytearray(b'\x01\x02\x03')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.reverse()
            + Opcode.REVERSEITEMS
            + Opcode.LDLOC0
            + Opcode.RET        # return a
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayReverse.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_extend(self):
        path = '%s/boa3_test/test_sc/bytes_test/BytearrayExtend.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_byte_array_to_int(self):
        data = b'\x01\x02'
        expected_output = (
            Opcode.PUSHDATA1    # bytearray(b'\x01\x02')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # bytearray(b'\x01\x02').to_int()
            + Type.int.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayToInt.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_to_int_with_builtin(self):
        data = b'\x01\x02'
        expected_output = (
            Opcode.PUSHDATA1    # bytearray(b'\x01\x02')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # bytearray.to_int(bytearray(b'\x01\x02'))
            + Type.int.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayToIntWithBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_byte_array_to_int_with_bytes_builtin(self):
        data = b'\x01\x02'
        expected_output = (
            Opcode.PUSHDATA1    # bytearray(b'\x01\x02')
            + Integer(len(data)).to_byte_array(min_length=1)
            + data
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.CONVERT    # bytes.to_int(bytearray(b'\x01\x02'))
            + Type.int.stack_item
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/bytes_test/BytearrayToIntWithBytesBuiltin.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)



import hashlib

from boa3.neo3.contracts import manifest
from boa3.neo3.core import serialization, IClonable, utils, types
from boa3.neo3.core.serialization import BinaryReader, BinaryWriter


class ContractState(serialization.ISerializable, IClonable):
    def __init__(self, script: bytes = None, _manifest: manifest.ContractManifest = None):
        self.script = script if script else b''
        self.manifest = _manifest if _manifest else manifest.ContractManifest()

    def __len__(self):
        return utils.get_var_size(self.script) + len(self.manifest)

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.script_hash() != other.script_hash():
            return False
        return True

    def serialize(self, writer: BinaryWriter) -> None:
        writer.write_var_bytes(self.script)
        writer.write_serializable(self.manifest)

    def deserialize(self, reader: BinaryReader) -> None:
        self.script = reader.read_var_bytes()
        self.manifest = reader.read_serializable(manifest.ContractManifest)

    def from_replica(self, replica):
        super().from_replica(replica)
        self.script = replica.script
        self.manifest = replica.manifest

    def clone(self):
        return ContractState(self.script, self.manifest)

    def script_hash(self) -> types.UInt160:
        """ Get the script hash."""
        intermediate_data = hashlib.sha256(self.script).digest()
        data = hashlib.new('ripemd160', intermediate_data).digest()
        return types.UInt160(data=data)



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestAny(BoaTest):

    def test_any_variable_assignments(self):
        two = String('2').to_bytes()
        path = '%s/boa3_test/test_sc/any_test/AnyVariableAssignments.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1      # a = 1
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # a = '2'
            + Integer(len(two)).to_byte_array() + two
            + Opcode.STLOC0
            + Opcode.PUSH1      # a = True
            + Opcode.STLOC0
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_variable_assignment_with_any(self):
        path = '%s/boa3_test/test_sc/any_test/VariableAssignmentWithAny.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_any_list(self):
        ok = String('ok').to_bytes()
        path = '%s/boa3_test/test_sc/any_test/AnyList.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_tuple(self):
        ok = String('ok').to_bytes()
        path = '%s/boa3_test/test_sc/any_test/AnyTuple.py' % self.dirname

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_operation(self):
        path = '%s/boa3_test/test_sc/any_test/OperationWithAny.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_function_any_param(self):
        ok = String('ok').to_bytes()
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT   # Main
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0          # bool_tuple = True, False
            + Opcode.PUSH1
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0         # SequenceFunction(bool_tuple)
            + Opcode.CALL
            + Integer(46).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction([True, 1, 'ok'])
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(36).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction('some_string')
            + Integer(len(some_string)).to_byte_array()
            + some_string
            + Opcode.CALL
            + Integer(21).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction((True, 1, 'ok'))
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(11).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH3          # SequenceFunction([1, 2, 3])
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(4).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHNULL
            + Opcode.RET        # return
            + Opcode.INITSLOT   # SequenceFunction
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0         # a = sequence
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/any_test/FunctionAnyParam.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_sequence_assignments(self):
        ok = String('ok').to_bytes()
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x05'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.PUSHDATA1  # any_tuple = (True, 1, 'ok')
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.PUSH0      # bool_tuple = True, False
            + Opcode.PUSH1
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.LDLOC0     # a = any_list
            + Opcode.STLOC4
            + Opcode.LDLOC2     # a = any_tuple
            + Opcode.STLOC4
            + Opcode.PUSHDATA1  # a = 'some_string'
            + Integer(len(some_string)).to_byte_array() + some_string
            + Opcode.STLOC4
            + Opcode.LDLOC1     # a = int_list
            + Opcode.STLOC4
            + Opcode.LDLOC3     # a = bool_tuple
            + Opcode.STLOC4
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/any_test/AnySequenceAssignments.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_int_sequence_any_assignments(self):
        path = '%s/boa3_test/test_sc/any_test/IntSequenceAnyAssignment.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_str_sequence_any_assignments(self):
        path = '%s/boa3_test/test_sc/any_test/StrSequenceAnyAssignment.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_str_sequence_str_assignment(self):
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # str_sequence = 'some_string'
            + Integer(len(some_string)).to_byte_array() + some_string
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/any_test/StrSequenceStrAssignment.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_any_sequence(self):
        ok = String('ok').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x05'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.PUSHDATA1  # any_tuple = (None, 1, 'ok')
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHNULL
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.PUSH0      # bool_tuple = True, False
            + Opcode.PUSH1
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.LDLOC3     # a = [any_list, int_list, any_tuple, bool_tuple]
            + Opcode.LDLOC2
            + Opcode.LDLOC1
            + Opcode.LDLOC0
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.STLOC4
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/any_test/SequenceOfAnySequence.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_int_sequence_success(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH8      # int_tuple = 10, 9, 8
            + Opcode.PUSH9
            + Opcode.PUSH10
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # a = [int_list, int_tuple]
            + Opcode.LDLOC0
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/any_test/SequenceOfIntSequence.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_int_sequence_fail(self):
        path = '%s/boa3_test/test_sc/any_test/SequenceOfAnyIntSequence.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



from boa3.boa3 import Boa3
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItemType import StackItemType
from boa3_test.tests.boa_test import BoaTest


class TestAssert(BoaTest):

    def test_assert_unary_boolean_operation(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0     # assert not a
            + Opcode.NOT
            + Opcode.ASSERT
            + Opcode.LDARG1     # return b
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertUnaryOperation.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_binary_boolean_operation(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0     # assert a != b
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertBinaryOperation.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_with_message(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a > 0, 'a must be greater than zero'
            + Opcode.PUSH0
            + Opcode.GT
            + Opcode.ASSERT         # neo assert doesn't receive messages yet
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertWithMessage.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_int(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertInt.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_str(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertStr.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_bytes(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertBytes.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_list(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.LDARG0     # return len(a)
            + Opcode.SIZE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertList.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_dict(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.LDARG0     # return len(a)
            + Opcode.SIZE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertDict.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assert_any(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Array
            + Opcode.JMPIF + Integer(12).to_byte_array(signed=True, min_length=1)
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Map
            + Opcode.JMPIF + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Struct
            + Opcode.JMPIFNOT + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/assert_test/AssertAny.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)



import hashlib

import ecdsa as _ecdsa  # type: ignore
from bitcoin import (  # type: ignore
    decompress as bitcoin_decompress,
    change_curve as bitcoin_change_curve,
    ecdsa_raw_sign as bitcoin_ecdsa_raw_sign)

from .bloomfilter import BloomFilter
from .ecc import (EllipticCurve, ECDSA)  # type: ignore
from .keypair import KeyPair
from .merkletree import MerkleTree

__all__ = ['EllipticCurve', 'ECDSA', 'MerkleTree', 'BloomFilter', 'KeyPair']


def verify_signature(message: bytes, signature: bytes, encoded_pub_key_bytes) -> bool:
    length = len(encoded_pub_key_bytes)
    if length != 33 and (encoded_pub_key_bytes[0] == 0x2 or encoded_pub_key_bytes[0] == 0x3):
        pubkey = bitcoin_decompress(encoded_pub_key_bytes)
    elif length == 65 and encoded_pub_key_bytes[0] == 0x4:
        pubkey = encoded_pub_key_bytes[1:]
    elif length != 64:
        raise ValueError("Invalid public key bytes")

    # TODO: replace the _ecdsa module usage with verifications from own ECC lib.
    #  For time reasons use the old 2.x tested logic
    # note that NIST256P == secp256r1, so we should be able to use ECDSA.secp256r1().verify() or re-use the bitcoin
    # package. There is also an bitcoin.ecsa_raw_verify that probably can work.
    try:
        vk = _ecdsa.VerifyingKey.from_string(pubkey, curve=_ecdsa.NIST256p, hashfunc=hashlib.sha256)
        res = vk.verify(signature, message, hashfunc=hashlib.sha256)
        return res
    except Exception:
        pass
    return False


def sign(message: bytes, private_key: bytes) -> bytes:
    bitcoin_change_curve(
        115792089210356248762697446949407573530086143415290314195533631308867097853951,
        115792089210356248762697446949407573529996955224135760342422259061068512044369,
        115792089210356248762697446949407573530086143415290314195533631308867097853948,
        41058363725152142129326129780047268409114441015993725554835256314039467401291,
        48439561293906451759052585252797914202762949526041747995844080717082404635286,
        36134250956749795798585127919587881956611106672985015071877198253568414405109
    )
    hash_ = hashlib.sha256(message).hexdigest()
    v, r, s = bitcoin_ecdsa_raw_sign(hash_, private_key)
    rb = r.to_bytes(32, 'big')
    sb = s.to_bytes(32, 'big')

    sig = rb + sb
    return sig



from __future__ import annotations

import hashlib
from enum import IntFlag
from typing import List

from boa3.neo3.core import serialization, utils, types, cryptography, Size as s
from boa3.neo3.network import payloads


class Cosigner(serialization.ISerializable):
    """
    A class that specifies who can pass CheckWitness() verifications in a smart contract.
    """

    def __init__(self, account: types.UInt160 = None,
                 scope: payloads.WitnessScope = None,
                 allowed_contracts: List[types.UInt160] = None,
                 allowed_groups: List[cryptography.EllipticCurve.ECPoint] = None):
        #: The TX sender.
        self.account = account if account else types.UInt160.zero()
        #: payloads.WitnessScope: The configured validation scope.
        self.scope = scope if scope else payloads.WitnessScope.GLOBAL
        #: List[types.UInt160]: Whitelist of contract hashes if used with
        #: :const:`~neo3.network.payloads.verification.WitnessScope.CUSTOM_CONTRACTS`.
        self.allowed_contracts = allowed_contracts if allowed_contracts else []
        #: List[cryptography.EllipticCurve.ECPoint]: Whitelist of public keys if used with
        #: :const:`~neo3.network.payloads.verification.WitnessScope.CUSTOM_GROUPS`.
        self.allowed_groups = allowed_groups if allowed_groups else []

    def __len__(self):
        contracts_size = 0
        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            contracts_size = utils.get_var_size(self.allowed_contracts)

        groups_size = 0
        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            groups_size = utils.get_var_size(self.allowed_groups)

        return s.uint160 + s.uint8 + contracts_size + groups_size

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.account)
        writer.write_uint8(self.scope)

        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            writer.write_serializable_list(self.allowed_contracts)

        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            writer.write_serializable_list(self.allowed_groups)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.account = reader.read_serializable(types.UInt160)
        self.scope = payloads.WitnessScope(reader.read_uint8())

        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            self.allowed_contracts = reader.read_serializable_list(types.UInt160)

        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            self.allowed_groups = reader.read_serializable_list(cryptography.EllipticCurve.ECPoint)


class Witness(serialization.ISerializable):
    """
    An executable verification script that validates a verifiable object like a transaction.
    """

    def __init__(self, invocation_script: bytes = None, verification_script: bytes = None):
        #: A set of VM instructions to setup the stack for verification.
        self.invocation_script = invocation_script if invocation_script else b''
        #: A set of VM instructions that does the actual verification.
        #: It is expected to set the result stack to a boolean True if validation passed.
        self.verification_script = verification_script if verification_script else b''
        self._script_hash = None

    def __len__(self):
        return utils.get_var_size(self.invocation_script) + utils.get_var_size(self.verification_script)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_bytes(self.invocation_script)
        writer.write_var_bytes(self.verification_script)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.invocation_script = reader.read_var_bytes(max=664)
        self.verification_script = reader.read_var_bytes(max=360)

    def script_hash(self) -> types.UInt160:
        """ Get the script hash based on the verification script."""
        intermediate_data = hashlib.sha256(self.verification_script).digest()
        data = hashlib.new('ripemd160', intermediate_data).digest()
        return types.UInt160(data=data)


class WitnessScope(IntFlag):
    """
    Determine the rules for a smart contract :func:`CheckWitness()` sys call.
    """
    #: Allow the witness in all context. Equal to NEO 2.x's default behaviour.
    GLOBAL = 0x00
    #: Allow the witness if the current calling script hash equals the entry script hash into the virtual machine.
    #: Using this prevents passing :func:`CheckWitness()` in a smart contract called via another smart contract.
    CALLED_BY_ENTRY = 0x01
    #: Allow the witness if called from a smart contract that is whitelisted in the cosigner
    #: :attr:`~neo3.network.payloads.verification.Cosigner.allowed_contracts` attribute.
    CUSTOM_CONTRACTS = 0x10
    #: Allow the witness if it any public key in the cosigner
    #: :attr:`~neo3.network.payloads.verification.Cosigner.allowed_groups` attribute is whitelisted in the contracts
    #: manifest.groups array.
    CUSTOM_GROUPS = 0x20



from __future__ import annotations

from contextlib import suppress
from copy import deepcopy
from typing import Iterator, Tuple, Dict

from boa3.neo3 import storage
from boa3.neo3.core import types
from boa3.neo3.network import payloads


class MemoryDB(storage.IDBImplementation):
    BLOCK = 'blocks'
    BLOCK_HEIGHT_MAP = 'blocksmap'
    BLOCK_BEST_HEIGHT = 'blockheight'
    CONTRACT = 'contracts'
    STORAGE = 'storages'
    TX = 'transactions'

    def __init__(self, options: dict = None):
        self.db: Dict[str, dict] = {
            self.BLOCK: {},
            self.BLOCK_HEIGHT_MAP: {},
            self.CONTRACT: {},
            self.STORAGE: {},
            self.TX: {}
        }
        self._best_block_height = -1

    def get_snapshotview(self) -> MemorySnapshot:
        return MemorySnapshot(self)

    def _internal_bestblockheight_get(self):
        if self._best_block_height == -1:
            raise KeyError
        return self._best_block_height

    def _internal_bestblockheight_put(self, height: int, batch=None):
        if batch:
            batch.put(self.BLOCK_BEST_HEIGHT, height, height)
        else:
            self._best_block_height = height

    def _internal_bestblockheight_update(self, height: int, batch=None):
        self._internal_bestblockheight_put(height, batch)

    def _internal_block_put(self, block: payloads.Block, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.BLOCK, block.hash(), block)
        else:
            self.db[self.BLOCK][block.hash()] = block
            self.db[self.BLOCK_HEIGHT_MAP][block.index] = block.hash()

        stored_value = -1
        with suppress(KeyError):
            stored_value = self._internal_bestblockheight_get()

        if block.index > stored_value:
            self._best_block_height = block.index

    def _internal_block_update(self, block: payloads.Block, batch: WriteBatch = None) -> None:
        self._internal_block_put(block, batch)

    def _internal_block_delete(self, hash: types.UInt256, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.BLOCK, hash)
        else:
            with suppress(KeyError):
                block = self.db[self.BLOCK].pop(hash)
                self.db[self.BLOCK_HEIGHT_MAP].pop(block.index)

    def _internal_block_get(self, hash: types.UInt256) -> payloads.Block:
        value = self.db[self.BLOCK].get(hash, None)
        if value is None:
            raise KeyError
        return deepcopy(value)

    def _internal_block_get_by_height(self, height: int) -> payloads.Block:
        block_hash = self.db[self.BLOCK_HEIGHT_MAP].get(height, None)
        if block_hash is None:
            raise KeyError

        return self._internal_block_get(block_hash)

    def _internal_block_all(self) -> Iterator[payloads.Block]:
        for block in self.db[self.BLOCK].values():
            yield deepcopy(block)

    def _internal_contract_put(self, contract: storage.ContractState, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.CONTRACT, contract.script_hash(), contract)
        else:
            self.db[self.CONTRACT][contract.script_hash()] = contract

    def _internal_contract_update(self, contract: storage.ContractState, batch: WriteBatch = None) -> None:
        self._internal_contract_put(contract, batch)

    def _internal_contract_delete(self, script_hash: types.UInt160, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.CONTRACT, script_hash)
        else:
            with suppress(KeyError):
                self.db[self.CONTRACT].pop(script_hash)

    def _internal_contract_get(self, script_hash: types.UInt160) -> storage.ContractState:
        value = self.db[self.CONTRACT].get(script_hash, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_contract_all(self) -> Iterator[storage.ContractState]:
        for contract in self.db[self.CONTRACT].values():
            yield deepcopy(contract)

    def _internal_storage_put(self, key: storage.StorageKey,
                              value: storage.StorageItem,
                              batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.STORAGE, key, value)
        else:
            self.db[self.STORAGE][key] = value

    def _internal_storage_update(self, key: storage.StorageKey,
                                 value: storage.StorageItem,
                                 batch: WriteBatch = None) -> None:
        self._internal_storage_put(key, value, batch)

    def _internal_storage_delete(self, key: storage.StorageKey, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.STORAGE, key)
        else:
            with suppress(KeyError):
                self.db[self.STORAGE].pop(key)

    def _internal_storage_get(self, key: storage.StorageKey) -> storage.StorageItem:
        value = self.db[self.STORAGE].get(key, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_storage_all(self, contract_script_hash: types.UInt160 = None) -> Iterator[Tuple[storage.StorageKey,
                                                                                                  storage.StorageItem]]:
        for k, v in self.db[self.STORAGE].items():
            if contract_script_hash:
                if contract_script_hash == k.contract:
                    yield deepcopy(k), deepcopy(v)
            else:
                yield deepcopy(k), deepcopy(v)

    def _internal_storage_find(self, contract_script_hash: types.UInt160,
                               key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        script_hash_len = 20
        for k, v in self.db[self.STORAGE].items():
            # k is of type StorageKey, which starts with a 20-byte script hash.
            # We skip this and search only in the `key` attribute
            if k.to_array()[script_hash_len:].startswith(key_prefix):
                yield deepcopy(k), deepcopy(v)

    def _internal_transaction_put(self, transaction: payloads.Transaction, batch: WriteBatch = None) -> None:
        if batch:
            batch.put(self.TX, transaction.hash(), transaction)
        else:
            self.db[self.TX][transaction.hash()] = transaction

    def _internal_transaction_update(self, transaction: payloads.Transaction, batch: WriteBatch = None) -> None:
        self._internal_transaction_put(transaction, batch)

    def _internal_transaction_delete(self, hash: types.UInt256, batch: WriteBatch = None) -> None:
        if batch:
            batch.delete(self.TX, hash)
        else:
            with suppress(KeyError):
                self.db[self.TX].pop(hash)

    def _internal_transaction_get(self, hash: types.UInt256) -> payloads.Transaction:
        value = self.db[self.TX].get(hash, None)
        if value is None:
            raise KeyError

        return deepcopy(value)

    def _internal_transaction_all(self) -> Iterator[payloads.Transaction]:
        for tx in self.db[self.TX].values():
            yield deepcopy(tx)

    def write_batch(self, batch) -> None:
        for table, action, pair in batch:
            if action == 'delete':
                item = self.db[table].pop(pair[0])
                if table == self.BLOCK:
                    self.db[self.BLOCK_HEIGHT_MAP].pop(item.index)
            elif action in ['update', 'add']:
                if table == self.BLOCK_BEST_HEIGHT:
                    self._best_block_height = pair[0]
                    continue
                self.db[table][pair[0]] = pair[1]
                if table == self.BLOCK:
                    # pair = (UInt256, Block)
                    self.db[self.BLOCK_HEIGHT_MAP][pair[1].index] = pair[0]


class WriteBatch:
    def __init__(self):
        self.statements = []

    def __iter__(self):
        for s in self.statements:
            yield s

    def put(self, table, key, value) -> None:
        self.statements.append((table, 'add', (key, value)))

    def delete(self, table, key) -> None:
        self.statements.append((table, 'delete', (key, None)))


class MemorySnapshot(storage.Snapshot):
    def __init__(self, db: MemoryDB):
        super(MemorySnapshot, self).__init__()
        self._db = db
        self._batch = WriteBatch()

        self._block_cache = MemoryDBCachedBlockAccess(db, self._batch)
        self._contract_cache = MemoryDBCachedContractAccess(db, self._batch)
        self._storage_cache = MemoryDBCachedStorageAccess(db, self._batch)
        self._tx_cache = MemoryDBCachedTXAccess(db, self._batch)
        self._block_height_cache = MemoryBestBlockHeightAttribute(db, self._batch)

    def commit(self) -> None:
        super(MemorySnapshot, self).commit()
        self._db.write_batch(self._batch)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # nothing to do
        pass


class MemoryBestBlockHeightAttribute(storage.AttributeCache):
    def __init__(self, db, batch):
        super(MemoryBestBlockHeightAttribute, self).__init__()
        self._db = db
        self._batch = batch

    def _get_internal(self):
        return self._db._internal_bestblockheight_get()

    def _update_internal(self, value):
        self._db._internal_bestblockheight_update(value, self._batch)


class MemoryDBCachedBlockAccess(storage.CachedBlockAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedBlockAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        for trackable in self._dictionary.values():  # trackable.item: payloads.Block
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_block_put(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_block_update(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_block_delete(trackable.item.hash(), self._batch)

    def create_snapshot(self):
        return storage.CloneBlockCache(self._db, self)


class MemoryDBCachedContractAccess(storage.CachedContractAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedContractAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        for trackable in self._dictionary.values():  # trackable.item: storage.ContractState
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_contract_put(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_contract_update(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_contract_delete(trackable.item.script_hash(), self._batch)

    def create_snapshot(self):
        return storage.CloneContractCache(self._db, self)


class MemoryDBCachedStorageAccess(storage.CachedStorageAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedStorageAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        for trackable in self._dictionary.values():
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_storage_put(trackable.key, trackable.item, self._batch)
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_storage_update(trackable.key, trackable.item, self._batch)
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_storage_delete(trackable.key, self._batch)

    def create_snapshot(self):
        return storage.CloneStorageCache(self._db, self)


class MemoryDBCachedTXAccess(storage.CachedTXAccess):
    def __init__(self, db, batch):
        super(MemoryDBCachedTXAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        for trackable in self._dictionary.values():  # trackable.item: payloads.Transaction
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_transaction_put(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_transaction_update(trackable.item, self._batch)
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_transaction_delete(trackable.item.hash(), self._batch)

    def create_snapshot(self):
        return storage.CloneTXCache(self._db, self)
